/**
 * Imports
 */

'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _bitVector = require('bit-vector');

/**
 * Actions
 */

var CREATE = 0;
var UPDATE = 1;
var MOVE = 2;
var REMOVE = 3;

/**
 * dift
 */

function dift(prev, next, effect) {
  var key = arguments.length <= 3 || arguments[3] === undefined ? defaultKey : arguments[3];

  var prevLen = prev.length;
  var nextLen = next.length;

  if (prevLen === 0) {
    if (nextLen === 0) return;else {
      // All creates
      for (var i = 0; i < nextLen; i++) {
        effect(CREATE, null, next[i], i);
      }
      return;
    }
  } else if (nextLen === 0) {
    // All removes
    for (var i = 0; i < prevLen; i++) {
      effect(REMOVE, prev[i], null, i);
    }

    return;
  }

  var pStartIdx = 0;
  var pEndIdx = prevLen - 1;
  var nStartIdx = 0;
  var nEndIdx = nextLen - 1;
  var pStartItem = prev[pStartIdx];
  var pEndItem = prev[pEndIdx];
  var nStartItem = next[nStartIdx];
  var nEndItem = next[nEndIdx];
  var created = 0;

  // List head is the same
  while (pStartIdx < prevLen && nStartIdx < nextLen && equal(pStartItem, nStartItem)) {
    effect(UPDATE, pStartItem, nStartItem);
    pStartItem = prev[++pStartIdx];
    nStartItem = next[++nStartIdx];
  }

  // Reversed
  while (pStartIdx <= pEndIdx && nStartIdx <= nEndIdx && equal(pStartItem, nEndItem)) {
    effect(MOVE, pStartItem, nEndItem, nEndIdx);
    pStartItem = prev[++pStartIdx];
    nEndItem = next[--nEndIdx];
  }

  // The above two cases each could have processed the entire list (whereas the next two
  // cannot if these didn't).  So bail out early here if we can.
  if (nStartIdx === nextLen && pStartIdx === prevLen) {
    return;
  }

  // Reversed the other way (in case of e.g. reverse and append)
  while (pEndIdx >= pStartIdx && nStartIdx <= nEndIdx && equal(nStartItem, pEndItem)) {
    effect(MOVE, pEndItem, nStartItem, nStartIdx);
    pEndItem = prev[--pEndIdx];
    nStartItem = next[++nStartIdx];
  }

  // List tail is the same
  while (pEndIdx >= pStartIdx && nEndIdx >= nStartIdx && equal(pEndItem, nEndItem)) {
    effect(UPDATE, pEndItem, nEndItem);
    pEndItem = prev[--pEndIdx];
    nEndItem = next[--nEndIdx];
  }

  var prevMap = keyMap(prev, pStartIdx, pEndIdx + 1, key);
  var keepBase = pStartIdx;
  var keep = (0, _bitVector.createBv)(pEndIdx - pStartIdx);

  for (; nStartIdx <= nEndIdx; nStartItem = next[++nStartIdx]) {
    var oldIdx = prevMap[key(nStartItem)];

    if (isUndefined(oldIdx)) {
      effect(CREATE, null, nStartItem, nStartIdx);
      ++created;
    } else {
      (0, _bitVector.setBit)(keep, oldIdx - keepBase);
      effect(oldIdx === nStartIdx ? UPDATE : MOVE, prev[oldIdx], nStartItem, nStartIdx);
    }
  }

  // If there are no creations, then you have to
  // remove exactly max(prevLen - nextLen, 0) elements in this
  // diff. You have to remove one more for each element
  // that was created. This means once we have
  // removed that many, we can stop.
  var necessaryRemovals = prevLen - nextLen + created;
  for (var removals = 0; removals < necessaryRemovals; pStartItem = prev[++pStartIdx]) {
    if (!(0, _bitVector.getBit)(keep, pStartIdx - keepBase)) {
      effect(REMOVE, pStartItem);
      ++removals;
    }
  }

  function equal(a, b) {
    return key(a) === key(b);
  }
}

function defaultKey(a) {
  return a.key;
}

function isUndefined(val) {
  return typeof val === 'undefined';
}

function keyMap(items, start, end, key) {
  var map = {};

  for (var i = start; i < end; ++i) {
    map[key(items[i])] = i;
  }

  return map;
}

/**
 * Exports
 */

exports['default'] = dift;
exports.CREATE = CREATE;
exports.UPDATE = UPDATE;
exports.MOVE = MOVE;
exports.REMOVE = REMOVE;